/*
 *      @(#)PickDevice.java 1.1 99/07/28 12:54:05
 *
 * Copyright (c) 1996-1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Sun grants you ("Licensee") a non-exclusive, royalty free, license to use,
 * modify and redistribute this software in source and binary code form,
 * provided that i) this copyright notice and license appear on all copies of
 * the software; and ii) Licensee does not utilize the software in a manner
 * which is disparaging to Sun.
 *
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY
 * IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR
 * NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN AND ITS LICENSORS SHALL NOT BE
 * LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THE SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL SUN OR ITS
 * LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF
 * OR INABILITY TO USE SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 * This software is not designed or intended for use in on-line control of
 * aircraft, air traffic, aircraft navigation or aircraft communications; or in
 * the design, construction, operation or maintenance of any nuclear
 * facility. Licensee represents and warrants that it will not use or
 * redistribute the Software for such purposes.
 */

import javax.media.j3d.*;
import javax.vecmath.*;
import java.awt.Point;
import java.awt.event.MouseEvent;
import java.text.NumberFormat;
import java.util.*;
import pickingExt.PickObject;

/**
 * An object to do picking and highlighting. 
 */

public class PickDevice {

    BranchGroup 	scene;
    Appearance  	highlightAppearance;
    PickObject		pickObject;
    Vector		pickedShapes = new Vector();
    Vector		unhighlightShapes = new Vector();
    Vector		highlightShapes = new Vector();
    Hashtable		savedAppearances = new Hashtable();

    MouseEvent		curPickEvent = null;

    NumberFormat 	nf;

    static final boolean debug = false;
    static final boolean timing = false;

    int			numPicks = 0;
    long		pickTime = 0;

    public PickDevice(Canvas3D canvas, BranchGroup scene,
	    Appearance highlightAppearance) {
	this.scene = scene;

        // set up a NumberFormat
        nf = NumberFormat.getInstance();
        nf.setMaximumFractionDigits(1);
        nf.setGroupingUsed(false);

	// set up the pick object
	pickObject = new PickObject(canvas, scene);
	int size = 4;
	// make a diamond shaped pick aperature
	Point[] pickAperture = new Point[4];
	pickAperture[0] = new Point(-size,     0);
	pickAperture[1] = new Point(    0, -size);
	pickAperture[2] = new Point( size,     0);
	pickAperture[3] = new Point(    0,  size);
	pickObject.setPickAperture(pickAperture);
	pickObject.setPickShapeMode(PickObject.SHAPE_RAY);
	//pickObject.setPickShapeMode(PickObject.SHAPE_RAY_APERTURE);
	//pickObject.setPickMode(PickObject.USE_GEOMETRY);

	this.highlightAppearance = highlightAppearance;

    }

    private void resetTiming() {
	numPicks = 0;
	pickTime = 0;
    }

    // unhighlight (revert to normal appearance) the shapes on the unhighlight
    // list
    private int unhighlightShapes() {
	int count = 0;
	for (Enumeration e = unhighlightShapes.elements(); 
			e.hasMoreElements(); ) {
	    Shape3D shape = (Shape3D)e.nextElement();
	    Appearance appearance = (Appearance)savedAppearances.remove(shape);
	    shape.setAppearance(appearance);
	}
	unhighlightShapes.clear();
	if (debug) {
	    System.out.println("unhighlight " + count + " shapes");
	}
	return count;
    }

    // highlight (change to the highlight appearance) the shapes on the 
    // highlight list
    private int highlightShapes() {
	int count = 0;
	for (Enumeration e = highlightShapes.elements(); e.hasMoreElements();){
	    Shape3D shape = (Shape3D)e.nextElement();
	    Appearance appearance = shape.getAppearance();
	    savedAppearances.put(shape, appearance);
	    shape.setAppearance(highlightAppearance);
	    count++;
	}
	highlightShapes.clear();
	if (debug) {
	    System.out.println("highlight " + count + " shapes");
	}
	return count;
    }

    // Swap the unhighlight list with the picked list. The unhighlight list
    // should be empty
    private void swapPickedShapes() {
	if (unhighlightShapes.size() > 0) {
	    System.out.println(
		"Warning, unhighlight shapes list was not cleared");
	}
	Vector temp = unhighlightShapes;
	unhighlightShapes = pickedShapes;
	pickedShapes = temp;
    }

    // add a shape to the picked list.  If it was previously highlighted, 
    // remove it from the unhighlight list, else add it to the highlight list
    private void addPickedShape(Shape3D pickedShape) {
	boolean found = false;
	int numUnhighlight = unhighlightShapes.size();
	for (int i = 0; (!found) && (i < numUnhighlight); i++) {
	    Shape3D shape = (Shape3D) unhighlightShapes.elementAt(i);
	    if (shape == pickedShape) {
		if (debug) {
		    System.out.println("found already highlit shape: " + shape);
		}
		pickedShape = shape;
		found = true;
		unhighlightShapes.remove(i);
	    }
	}
	if (!found) {
	    if (debug) {
		System.out.println("Shape to highlight:"+pickedShape);
	    }
	    highlightShapes.add(pickedShape);
	}
	pickedShapes.add(pickedShape);
    }

    public void setPickEvent(MouseEvent pickEvent) {
	curPickEvent = pickEvent;
    }

    // Do a pick/highlight for the current pick position
    public void processPickEvent() {

	if (curPickEvent == null) {
	    return;
	}

	int mouseX = curPickEvent.getX();
	int mouseY = curPickEvent.getY();

	long start = 0;
	if (timing) {
	    start = System.currentTimeMillis();
	}
	SceneGraphPath[] paths = pickObject.pickAll(mouseX, mouseY);
	if (timing) {
	    numPicks++;
	    pickTime += System.currentTimeMillis() - start;
	    if ((numPicks % 10) == 0) {
		System.out.println(numPicks + " picks in " + pickTime + 
			" ms, " + nf.format(pickTime/numPicks) + " ms/pick");
		resetTiming();
	    }
	}

	swapPickedShapes();
	     
	if (paths != null) { 
	    if (debug) {
		System.out.println(paths.length + " pick hit(s)");
	    }
	    for (int i = 0; i < paths.length; i++) {
		Shape3D pickedShape = (Shape3D) paths[i].getObject();
		addPickedShape(pickedShape);
	    }
	}  else {
	    if (debug) {
	        System.out.println("no pick hit");
	    }
	}

	int count = unhighlightShapes();
	count += highlightShapes();

	if (debug && (count > 0)) {
	    System.out.println("Edited " + count + " shapes");
	}
    }

    // unhighlight the currently picked shapes
    public void clearPickHighlights() {
	swapPickedShapes();
	unhighlightShapes();
    }


    /**
     * Get the shapes which were picked the last time processPickEvents was
     * called.  TODO: save complete pick paths.
     */
    Shape3D[] getPickedShapes() {
	int i = 0;
	Shape3D[] retval = new Shape3D[pickedShapes.size()];
	for (Enumeration e = pickedShapes.elements(); 
			e.hasMoreElements(); ) {
	    Shape3D shape = (Shape3D) e.nextElement();
	    if (!shape.isLive()) {
		shape = (Shape3D) shape.getUserData();
	    }
	    retval[i++] = shape;
	}
	return retval;
    }

}

