
/*
 *      @(#)PickFixShape3D.java 1.2 99/09/01 13:59:25
 *
 * Copyright (c) 1996-1999 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Sun grants you ("Licensee") a non-exclusive, royalty free, license to use,
 * modify and redistribute this software in source and binary code form,
 * provided that i) this copyright notice and license appear on all copies of
 * the software; and ii) Licensee does not utilize the software in a manner
 * which is disparaging to Sun.
 *
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY
 * IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR
 * NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN AND ITS LICENSORS SHALL NOT BE
 * LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THE SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL SUN OR ITS
 * LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF
 * OR INABILITY TO USE SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 * This software is not designed or intended for use in on-line control of
 * aircraft, air traffic, aircraft navigation or aircraft communications; or in
 * the design, construction, operation or maintenance of any nuclear
 * facility. Licensee represents and warrants that it will not use or
 * redistribute the Software for such purposes.
 */


import javax.media.j3d.*;
import javax.vecmath.*;

/**
 * A helper class which makes the Shape3D have a slightly larger extent
 * to make it possible to use a PickRay to pick shapes with a zero size
 * for one or more dimensions of their extent (i.e. a point).  The extent
 * will be made greater by the epsilon value in each direction when the
 * shape is created (if the geometry is specified), or when the
 * updateGeometry() method is called.
 */

public class PickFixShape3D extends Shape3D {
    double epsilon = 1e-02;

    public PickFixShape3D() {
	super();
    }

    public PickFixShape3D(Geometry geometry) {
	super(geometry);
	updateBounds(geometry);
    }

    public PickFixShape3D(Geometry geometry, Appearance appearance)
    {
	super(geometry, appearance);
	updateBounds(geometry);
    }

    /**
     * Sets the epsilon value for the bounding box for this prim
     */
    public void setEpsilon(double epsilon) {
	this.epsilon = epsilon;
    }

    /**
     * Gets the epsilon value for the bounding box for this prim
     */
    public double getEpsilon() {
	return epsilon;
    }


    /** 
     * Replacement for setGeometry(), which can't be overridden.  Sets
     * the geometry and then calls updateBounds. Must be called before
     * the Shape3D is live or cap bits need to be set
     */
    public void updateGeometry(Geometry geo) {
	setGeometry(geo);
	updateBounds(geo);
    }

    /**
     * Called after changing the geometry.  Must be called before Shape3D
     * is live or cap bits need to be set
     */
    public void updateBounds(Geometry geo) {
	if ((geo != null) && (geo instanceof GeometryArray)) {
	    GeometryArray ga = (GeometryArray) geo;
	    int numVerts = ga.getVertexCount();
	    if (numVerts > 0) {
		// a flat array if fastest to work with 
		double[] coords = new double[numVerts * 3];
		ga.getCoordinates(0, coords);
		Point3d min = new Point3d(coords[0], coords[1], coords[2]);
		Point3d max = new Point3d(coords[0], coords[1], coords[2]);
		for (int i = 1; i < numVerts; i++) {
		    double curX = coords[i * 3];
		    double curY = coords[i * 3 + 1];
		    double curZ = coords[i * 3 + 2];
		    if (curX < min.x) {
			min.x = curX;
		    }
		    if (curY < min.y) {
			min.y = curY;
		    }
		    if (curZ < min.z) {
			min.z = curZ;
		    }
		    if (curX > max.x) {
			max.x = curX;
		    }
		    if (curY > max.y) {
			max.y = curY;
		    }
		    if (curZ > max.z) {
			max.z = curZ;
		    }
		}
		// add an epsilon value to bounds
		min.x -= epsilon;
		min.y -= epsilon;
		min.z -= epsilon;
		max.x += epsilon;
		max.y += epsilon;
		max.z += epsilon;
		BoundingBox bounds = new BoundingBox(min, max);
		// update the shape 
		setBoundsAutoCompute(false);
		setBounds(bounds);
	    }
	}
    }
}
